#include "gmp-impl.h"
#include "longlong.h"

#if defined (__amd64__) && W_TYPE_SIZE == 64
#define CMP_SWAP(ap,bp,n)						\
  do {									\
    mp_limb_t tmp;							\
    __asm__ ("cmp\t%5, %6"	"\n\t"					\
	     "mov\t%1, %0"	"\n\t"					\
	     "cmovae\t%2, %1"	"\n\t"					\
	     "cmovae\t%0, %2"						\
	     : "=&r,&r" (tmp), "=r,r" (ap), "=r,r" (bp)			\
	     : "1,1" (ap), "2,2" (bp), "r,rm" (ap[n-1]), "rm,r" (bp[n-1])); \
  } while (0)
#endif

#ifndef CMP_SWAP
#define CMP_SWAP(ap,bp,n)						\
  do {									\
    mp_limb_t m = -(mp_limb_t) (ap[n - 1] < bp[n - 1]);			\
    mp_limb_t *_tp = ap;						\
    ap = (mp_ptr) (((size_t) ap & ~m) | ((size_t) bp & m));		\
    bp = (mp_ptr) (((size_t) bp & ~m) | ((size_t) _tp & m));		\
  } while (0)
#endif

#ifndef CMP_SWAP
#define CMP_SWAP(ap,bp,n)						\
  do {									\
    if (ap[n - 1] < bp[n - 1])						\
      {									\
	mp_limb_t *_tp = ap; ap = bp; bp = _tp;				\
      }									\
  } while (0)
#endif


#define SMASH(a,b) a##b

void mpn_gcd_33 (mp_limb_t *, mp_limb_t *, mp_limb_t *);
void mpn_gcd_44 (mp_limb_t *, mp_limb_t *, mp_limb_t *);
void mpn_gcd_55 (mp_limb_t *, mp_limb_t *, mp_limb_t *);

mp_limb_t mpn_sub_3 (mp_ptr, mp_srcptr, mp_srcptr);
mp_limb_t mpn_sub_4 (mp_ptr, mp_srcptr, mp_srcptr);
mp_limb_t mpn_sub_5 (mp_ptr, mp_srcptr, mp_srcptr);
void mpn_rshift_3 (mp_ptr, mp_srcptr, int);
void mpn_rshift_4 (mp_ptr, mp_srcptr, int);
void mpn_rshift_5 (mp_ptr, mp_srcptr, int);

static inline void
mpn_gcd_NN (mp_limb_t *rp, mp_limb_t *up, mp_limb_t *vp, size_t N)
{
  mp_limb_t cy;

  do
    {
      CMP_SWAP (up, vp, N);

#if HAVE_NATIVE_mpn_sub_3
      if (N == 3)
	cy = mpn_sub_3 (up, up, vp);
      else
#endif
#if HAVE_NATIVE_mpn_sub_4
      if (N == 4)
	cy = mpn_sub_4 (up, up, vp);
      else
#endif
#if HAVE_NATIVE_mpn_sub_5
      if (N == 5)
	cy = mpn_sub_5 (up, up, vp);
      else
#endif
	cy = mpn_sub_n (up, up, vp, N);

      if (UNLIKELY (cy))
	mpn_neg (up, up, N);

      if (UNLIKELY (up[0] == 0))
	{
	  // Handle any number of low U zeros including that U = 0
	  int all_zeros = 1;
	  mp_size_t i;
	  for (i = 1; i < N && all_zeros; i++)
	    all_zeros = (up[i] == 0);

	  if (all_zeros)
	    {
	      mpn_copyi (rp, vp, N);
	      return;
	    }
	  mpn_copyi (up, up + (i - 1), N - (i - 1));
	  mpn_zero (up + N - (i - 1), i - 1);
	  if ((up[0] & 1) != 0)
	    goto skip_shift;
	}
      int cnt;
      count_trailing_zeros (cnt, up[0]);
#if HAVE_NATIVE_mpn_rshift_3
      if (N == 3)
	mpn_rshift_3 (up, up, cnt);
      else
#endif
#if HAVE_NATIVE_mpn_rshift_4
      if (N == 4)
	mpn_rshift_4 (up, up, cnt);
      else
#endif
#if HAVE_NATIVE_mpn_rshift_5
      if (N == 5)
	mpn_rshift_5 (up, up, cnt);
      else
#endif
	mpn_rshift (up, up, N, cnt);
    skip_shift:;
    }
  while ((up[N - 1] | vp[N - 1]) != 0);

  switch (N)
    {
    case 3:
      {
	mp_double_limb_t g = mpn_gcd_22 (up[1], up[0], vp[1], vp[0]);
	rp[0] = g.d0;
	rp[1] = g.d1;
	rp[2] = 0;
	break;
      }
    case 4:
      mpn_gcd_33 (rp, up, vp);
      rp[3] = 0;
      break;
    case 5:
      mpn_gcd_44 (rp, up, vp);
      rp[4] = 0;
      break;
    }
}

void
SMASH(mpn_gcd_,33) (mp_limb_t *rp, mp_limb_t *up, mp_limb_t *vp)
{
  mpn_gcd_NN (rp, up, vp, 3);
}

void
SMASH(mpn_gcd_,44) (mp_limb_t *rp, mp_limb_t *up, mp_limb_t *vp)
{
  mpn_gcd_NN (rp, up, vp, 4);
}

void
SMASH(mpn_gcd_,55) (mp_limb_t *rp, mp_limb_t *up, mp_limb_t *vp)
{
  mpn_gcd_NN (rp, up, vp, 5);
}
